# SPDX-License-Identifier: CECILL-2.1
# Copyright (c) 2025 Synchrotron SOLEIL

"""
Plotting of particle trajectories and magnetic fields.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import scipy.integrate as integrate

from .style import start_plotting

# ---------------------------------------------------------------------------
# Electron trajectory
# ---------------------------------------------------------------------------

def plot_electron_trajectory(eBeamTraj: dict, direction: str, **kwargs) -> None:
    """
    Plot horizontal or vertical electron trajectory from simulation data.

    Args:
        eBeamTraj: Dictionary containing electron trajectory data.
        direction: 'horizontal', 'vertical', or 'both'.
        **kwargs: Optional keyword arguments, e.g., scaling factor `k`.
    """
    k = kwargs.get('k', 1)
    start_plotting(k)

    colors = ['darkred', 'olive', 'steelblue']
    s = eBeamTraj['eTraj']['Z']*1E3

    if direction.lower() in ['x', 'h', 'hor', 'horizontal']:
        B = eBeamTraj['eTraj']['By']
        graph_title = 'Horizontal electron trajectory'
        axis = 'X'
        label_B = 'By [T]'
    elif direction.lower() in ['y', 'v', 'ver', 'vertical']:
        B = eBeamTraj['eTraj']['Bx']
        graph_title = 'Vertical electron trajectory'
        axis = 'Y'
        label_B = 'Bx [T]'
    elif direction.lower() in ['b', 'both']:
        plot_electron_trajectory(eBeamTraj, 'horizontal', **kwargs)
        plot_electron_trajectory(eBeamTraj, 'vertical', **kwargs)
        return
    else:
        raise ValueError("Direction must be 'horizontal', 'vertical', or 'both'.")

    fig, axes = plt.subplots(3, 1, sharex=True, figsize=(8, 8), height_ratios=[1, 1, 1])
    fig.suptitle(graph_title, fontsize=16)
    fig.subplots_adjust(hspace=0.3)

    ys = [
        B,
        eBeamTraj['eTraj'][axis] * 1e3,
        eBeamTraj['eTraj'][f'{axis}p'] * 1e3
    ]

    labels = [label_B, f'{axis.lower()} [mm]', f"{axis.lower()}' [mrad]"]

    for ax, y, color, label in zip(axes, ys, colors, labels):
        ax.set_facecolor('white')
        ax.plot(s, y, color=color, linewidth=1.5, label=label)
        ax.grid(True, which='both', color='gray', linestyle=':', linewidth=0.5)
        ax.tick_params(direction='in', top=True, right=True)
        for spine in ['top', 'right', 'bottom', 'left']:
            ax.spines[spine].set_visible(True)
            ax.spines[spine].set_color('black')
        ax.yaxis.label.set_color('black')
        ax.legend(loc='upper right', frameon=True)

    axes[-1].set_xlabel('[mm]', color='black')
    plt.tight_layout()
    plt.show()

# ---------------------------------------------------------------------------
# Magnetic field
# ---------------------------------------------------------------------------

def plot_magnetic_field(eBeamTraj: dict, direction: str, **kwargs) -> None:
    """
    Plot horizontal or vertical magnetic field from simulation data.

    Args:
        eBeamTraj: Dictionary containing electron trajectory data.
        direction: 'horizontal', 'vertical', or 'both'.
        **kwargs: Optional keyword arguments, e.g., scaling factor `k`.
    """
    k = kwargs.get('k', 1)
    start_plotting(k)

    colors = ['darkred', 'olive', 'steelblue']
    s = eBeamTraj['eTraj']['Z']*1E3

    if direction.lower() in ['x', 'h', 'hor', 'horizontal']:
        B = eBeamTraj['eTraj']['By']
        graph_title = 'Vertical magnetic field'
    elif direction.lower() in ['y', 'v', 'ver', 'vertical']:
        B = eBeamTraj['eTraj']['Bx']
        graph_title = 'Horizontal magnetic field'
    elif direction.lower() in ['b', 'both']:
        plot_magnetic_field(eBeamTraj, 'horizontal', **kwargs)
        plot_magnetic_field(eBeamTraj, 'vertical', **kwargs)
        return
    else:
        raise ValueError("Direction must be 'horizontal', 'vertical', or 'both'.")

    fig, axes = plt.subplots(3, 1, sharex=True, figsize=(8, 8), height_ratios=[1, 1, 1])
    fig.suptitle(graph_title, fontsize=16)
    fig.subplots_adjust(hspace=0.3)

    frst_field_integral = integrate.cumulative_trapezoid(B, s, initial=0)
    scnd_field_integral = integrate.cumulative_trapezoid(frst_field_integral, s, initial=0)

    ys = [
        B,
        frst_field_integral,
        scnd_field_integral
    ]
    labels = ['B [T]', '$\int$B$\cdot$d$s$ [T$\cdot$m]', '$\iint$B$\cdot$d$s$² [T$\cdot$m²]']

    for ax, y, color, label in zip(axes, ys, colors, labels):
        ax.set_facecolor('white')
        ax.plot(s, y, color=color, linewidth=1.5, label=label)
        ax.grid(True, which='both', color='gray', linestyle=':', linewidth=0.5)
        ax.tick_params(direction='in', top=True, right=True)
        for spine in ['top', 'right', 'bottom', 'left']:
            ax.spines[spine].set_visible(True)
            ax.spines[spine].set_color('black')
        ax.yaxis.label.set_color('black')
        ax.legend(loc='upper right', frameon=True)

    axes[-1].set_xlabel('[mm]', color='black')
    plt.tight_layout()
    plt.show()

def plot_field_and_twiss(data: dict, **kwargs) -> None:
    """
    Plot magnetic field components vs s, and (optionally) Twiss parameters.

    Parameters
    ----------
    data : dict
        Required:
          - 's' : array, positions [m], shape (N,)
          - 'B' : array, magnetic field vectors [T], shape (N, 3) for (Bx, By, Bz)
        Optional:
          - 'Twiss' : dict with:
              * 'beta_x' : array, [m], shape (N,)
              * 'beta_y' : array, [m], shape (N,)
              * 'eta_x' or 'eta' : array, [m], shape (N,)
    **kwargs
        k : float, optional
            Global scaling factor for fonts (default matches aux_plots).

    Raises
    ------
    KeyError
        If required keys are missing in `data`.
    ValueError
        If array shapes are inconsistent or values are non-finite.
    """
    k = kwargs.get('k', 1)
    try:
        start_plotting(k)
    except NameError:
        pass

    if 's' not in data or 'B' not in data:
        raise KeyError("Missing required keys: 's' and 'B' must be present.")

    s = np.asarray(data['s'], dtype=float)
    B = np.asarray(data['B'], dtype=float)

    if s.ndim != 1:
        raise ValueError("'s' must be 1D.")
    if B.ndim != 2 or B.shape[1] != 3 or B.shape[0] != s.size:
        raise ValueError("'B' must have shape (N, 3) with N == len(s).")
    if not (np.isfinite(s).all() and np.isfinite(B).all()):
        raise ValueError("'s' and 'B' must be finite.")

    color_Bx = 'darkred'
    color_By = 'olive'
    color_Bz = 'steelblue'
    color_beta_x = 'teal'
    color_beta_y = 'peru'
    color_eta    = 'slategray'

    def style_axis(ax):
        ax.set_facecolor('white')
        ax.grid(True, which='both', color='gray', linestyle=':', linewidth=0.5)
        ax.tick_params(direction='in', top=True, right=True)
        for spine in ('top', 'right', 'bottom', 'left'):
            ax.spines[spine].set_visible(True)
            ax.spines[spine].set_color('black')

    tw = data.get('Twiss', None)
    has_twiss = tw is not None

    if has_twiss:
        for key in ('beta_x', 'beta_y'):
            if key not in tw:
                raise KeyError(f"Missing Twiss key: '{key}'.")
        eta_key = 'eta_x' if 'eta_x' in tw else ('eta' if 'eta' in tw else None)
        if eta_key is None:
            raise KeyError("Missing Twiss key: 'eta_x' (or 'eta').")

        beta_x = np.asarray(tw['beta_x'], dtype=float)
        beta_y = np.asarray(tw['beta_y'], dtype=float)
        eta    = np.asarray(tw[eta_key], dtype=float)
        for arr, name in ((beta_x, 'beta_x'), (beta_y, 'beta_y'), (eta, eta_key)):
            if arr.ndim != 1 or arr.size != s.size:
                raise ValueError(f"'{name}' must be 1D with len == len(s).")
            if not np.isfinite(arr).all():
                raise ValueError(f"'{name}' must be finite.")

        fig, (axB, axTw) = plt.subplots(
            nrows=2, ncols=1, sharex=True, figsize=(9, 7),
            gridspec_kw={'height_ratios': [1, 1]}
        )
        fig.subplots_adjust(hspace=0.25)
        fig.suptitle("Magnetic field and Twiss parameters vs s",
                     fontsize=16 * k, x=0.5)

        style_axis(axB)
        labels = []
        for y, lbl, col in ((B[:, 0], 'Bx', color_Bx),
                            (B[:, 1], 'By', color_By),
                            (B[:, 2], 'Bz', color_Bz)):
            if np.any(y != 0):
                axB.plot(s, y, color=col, lw=1.5, label=lbl)
                axB.fill_between(s, y, 0.0, facecolor=col, alpha=0.5)
                labels.append(lbl)
        if labels:
            axB.legend(loc='upper right', frameon=True)
        axB.set_ylabel('B [T]')

        style_axis(axTw)
        axTw.plot(s, beta_x, color=color_beta_x, lw=1.5, label=r'$\beta_x$')
        axTw.plot(s, beta_y, color=color_beta_y, lw=1.5, label=r'$\beta_y$')
        axTw.set_ylabel(r'$\beta$ [m]', color='black')
        axTw.legend(loc='upper left', frameon=True)

        axTw2 = axTw.twinx()
        axTw2.tick_params(direction='in', top=True, right=True)
        axTw2.plot(s, eta, color=color_eta, lw=1.5, label=r'$\eta_x$')
        axTw2.set_ylabel(r'$\eta_x$ [m]', color=color_eta)

        axTw.set_xlabel('s [m]')
        plt.tight_layout(rect=[0, 0, 1, 0.95])
        plt.show()

    else:
        fig, axB = plt.subplots(figsize=(9, 4.5))
        fig.suptitle("Magnetic field vs s", fontsize=16 * k, x=0.5)
        style_axis(axB)

        axB.plot(s, B[:, 0], color=color_Bx, lw=1.5, label='Bx')
        axB.fill_between(s, B[:, 0], 0.0, facecolor=color_Bx, alpha=0.5)
        axB.plot(s, B[:, 1], color=color_By, lw=1.5, label='By')
        axB.fill_between(s, B[:, 1], 0.0, facecolor=color_By, alpha=0.5)

        axB.set_ylabel('B [T]')
        axB.set_xlabel('s [m]')
        axB.legend(loc='upper right', frameon=True)

        plt.tight_layout()
        plt.show()


# SPDX-License-Identifier: CECILL-2.1
# Copyright (c) 2025 Synchrotron SOLEIL

"""
Module for processing wavefront data.
"""

from __future__ import annotations

from copy import copy

import numpy as np


def integrate_wavefront_window(wfrDict, hor_slit, ver_slit):
    """
    Integrates the wavefront intensity within a rectangular window.

    Parameters:
        wfrDict (dict): Dictionary containing wavefront data as produced by write_wavefront.
        hor_slit (tuple or float): (x_min, x_max) limits of window [m], or positive float interpreted as (-w/2, +w/2).
        ver_slit (tuple or float): (y_min, y_max) limits of window [m], or positive float interpreted as (-w/2, +w/2).

    Returns:
        dict: Dictionary with structure {polarisation: integrated_intensity, ...}, plus the window used.
    """

    x_axis = wfrDict['axis']['x']
    y_axis = wfrDict['axis']['y']

    if isinstance(hor_slit, (int, float)):
        if hor_slit <= 0:
            raise ValueError("hor_slit must be positive if passed as float.")
        half_w = hor_slit / 2
        hor_slit = (-half_w, half_w)

    if isinstance(ver_slit, (int, float)):
        if ver_slit <= 0:
            raise ValueError("ver_slit must be positive if passed as float.")
        half_w = ver_slit / 2
        ver_slit = (-half_w, +half_w)

    x_min, x_max = hor_slit
    y_min, y_max = ver_slit

    x_mask = (x_axis >= x_min) & (x_axis <= x_max)
    y_mask = (y_axis >= y_min) & (y_axis <= y_max)

    dx = (x_axis[1] - x_axis[0]) * 1e3  # mm
    dy = (y_axis[1] - y_axis[0]) * 1e3  # mm

    results = {'hor_slit': hor_slit, 'ver_slit': ver_slit}

    for pol in wfrDict['intensity']:
        map2d = copy(wfrDict['intensity'][pol])
        subarray = map2d[np.ix_(y_mask, x_mask)]
        integrated = np.sum(subarray) * dx * dy
        results[pol] = integrated

    return results


# SPDX-License-Identifier: CECILL-2.1
# Copyright (c) 2025 Synchrotron SOLEIL

"""
Plots for 2D spatial maps: wavefronts, power density, and CSD.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LogNorm
from skimage.restoration import unwrap_phase

from .style import igor_cmap, srw_cmap, start_plotting

# ---------------------------------------------------------------------------
# Wavefront
# ---------------------------------------------------------------------------

def plot_wavefront(
    wfr: dict,
    cuts: bool = True,
    show_phase: bool = True,
    log_intensity: bool = False,
    **kwargs,
) -> None:
    """
    Plot wavefront intensity and optionally the phase from a wavefront dictionary.

    For each polarisation in the dictionary, this plots:
        - If cuts=True: 2D intensity map + horizontal (y=0) and vertical (x=0) cuts.
        - If cuts=False: Only the 2D intensity map.
    Phase map is shown at the end if requested, with optional unwrapping.

    Intensity can be shown either in linear scale (default) or logarithmic scale
    using Matplotlib's log normalization for the 2D map and semilogy for the cuts.

    Parameters
    ----------
    wfr : dict
        Dictionary returned by `write_wavefront` or `read_wavefront`.
    cuts : bool, optional
        Whether to include 1D cuts in the plot (default: True).
    show_phase : bool, optional
        Whether to plot the phase maps (default: True).
    log_intensity : bool, optional
        If True, use log scale for intensity (2D via LogNorm, cuts via semilogy).
        Phase is always plotted in linear scale (default: False).
    **kwargs :
        k : float, optional
            Scaling factor for fonts and titles (default: 1).
        vmin : float, optional
            Minimum value for intensity color scale (linear units).
        vmax : float, optional
            Maximum value for intensity color scale (linear units).
        unwrap : bool, optional
            Whether to unwrap the phase before plotting (default: True).
        cmap : str, optional
            Colormap for intensity. Can be any Matplotlib cmap name (e.g. 'viridis'),
            or the special keywords:
                - 'srw'  : black → white
                - 'igor' : black → navy → darkred → red → orange → yellow
            Default is 'jet'.
    """
    k = kwargs.get("k", 1)
    vmin = kwargs.get("vmin", None)
    vmax = kwargs.get("vmax", None)
    unwrap = kwargs.get("unwrap", True)
    cmap_name = kwargs.get("cmap", "jet")

    if cmap_name == "srw":
        cmap_intensity = srw_cmap
    elif cmap_name == "igor":
        cmap_intensity = igor_cmap
    else:
        cmap_intensity = cmap_name 

    start_plotting(k)

    x = wfr["axis"]["x"] * 1e3  # mm
    y = wfr["axis"]["y"] * 1e3  # mm
    X, Y = np.meshgrid(x, y)

    dx = wfr["axis"]["x"][1] - wfr["axis"]["x"][0]
    dy = wfr["axis"]["y"][1] - wfr["axis"]["y"][0]

    fctr = (wfr["axis"]["x"][-1] - wfr["axis"]["x"][0]) / (
        wfr["axis"]["y"][-1] - wfr["axis"]["y"][0]
    )

    for pol, data in wfr["intensity"].items():
        flux = np.sum(data * dx * 1e3 * dy * 1e3)

        if log_intensity:
            data_masked = np.ma.masked_less_equal(data, 0.0)
            positive = data[data > 0]

            if positive.size == 0:
                continue

            vmin_eff = vmin if (vmin is not None and vmin > 0) else positive.min()
            vmax_eff = vmax if (vmax is not None and vmax > vmin_eff) else data.max()
            norm = LogNorm(vmin=vmin_eff, vmax=vmax_eff)
            vmin_lin = None
            vmax_lin = None
        else:
            data_masked = data
            norm = None
            vmin_lin = vmin
            vmax_lin = vmax

        fig = plt.figure(figsize=(4.2 * fctr, 4))
        fig.suptitle(
            f"({pol}) | flux: {flux:.2e} ph/s/0.1%bw",
            fontsize=16 * k,
            x=0.5,
        )
        ax = fig.add_subplot(111)

        im = ax.pcolormesh(
            X,
            Y,
            data_masked,
            shading="auto",
            cmap=cmap_intensity,
            vmin=vmin_lin,
            vmax=vmax_lin,
            norm=norm,
        )

        ax.set_aspect("equal")
        ax.set_xlabel("x [mm]")
        ax.set_ylabel("y [mm]")
        ax.grid(True, linestyle=":", linewidth=0.5)

        cb = plt.colorbar(im, ax=ax, fraction=0.046, pad=0.04)

        plt.show()

        if cuts:
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 4), sharey=True)
            ix0 = np.argmin(np.abs(x))
            iy0 = np.argmin(np.abs(y))

            hor = data[iy0, :]
            ver = data[:, ix0]

            if log_intensity:
                hor_plot = np.where(hor > 0, hor, np.nan)
                ver_plot = np.where(ver > 0, ver, np.nan)

                ax1.semilogy(x, hor_plot, color="darkred", lw=1.5)
                ax2.semilogy(y, ver_plot, color="darkred", lw=1.5)
            else:
                ax1.plot(x, hor, color="darkred", lw=1.5)
                ax2.plot(y, ver, color="darkred", lw=1.5)

            ax1.set_title("Hor. cut (y=0)")
            ax1.set_xlabel("x [mm]")
            ax1.set_ylabel(r"ph/s/mm$^2$/0.1%bw")
            ax1.grid(True, linestyle=":", linewidth=0.5)
            ax1.tick_params(direction="in", top=True, right=True)

            ax2.set_title("Ver. cut (x=0)")
            ax2.set_xlabel("y [mm]")
            ax2.grid(True, linestyle=":", linewidth=0.5)
            ax2.tick_params(direction="in", top=True, right=True)

            plt.tight_layout(rect=[0, 0, 1, 0.95])
            plt.show()

        if show_phase:
            phase = wfr["phase"][pol]
            if unwrap:
                phase = unwrap_phase(phase)
                phase -= phase[phase.shape[0] // 2, phase.shape[1] // 2]
                cmapref = "terrain"
            else:
                cmapref = "coolwarm"

            Rx = wfr["wfr"].Rx
            Ry = wfr["wfr"].Ry

            fig = plt.figure(figsize=(4.2 * fctr, 4))
            fig.suptitle(
                f"({pol}) | residual phase - Rx = {Rx:.2f} m, Ry = {Ry:.2f} m",
                fontsize=16 * k,
                x=0.5,
            )
            ax = fig.add_subplot(111)
            im = ax.pcolormesh(X, Y, phase, shading="auto", cmap=cmapref)
            ax.set_aspect("equal")
            ax.set_xlabel("x [mm]")
            ax.set_ylabel("y [mm]")
            ax.grid(True, linestyle=":", linewidth=0.5)
            cb = plt.colorbar(im, ax=ax, fraction=0.046, pad=0.04)
            plt.show()

            if cuts:
                fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 4), sharey=True)
                ix0 = np.argmin(np.abs(x))
                iy0 = np.argmin(np.abs(y))

                ax1.plot(x, phase[iy0, :], color="darkred", lw=1.5)
                ax1.set_title("Hor. cut (y=0)")
                ax1.set_xlabel("x [mm]")
                ax1.set_ylabel("rad")
                ax1.grid(True, linestyle=":", linewidth=0.5)
                ax1.tick_params(direction="in", top=True, right=True)

                ax2.plot(y, phase[:, ix0], color="darkred", lw=1.5)
                ax2.set_title("Ver. cut (x=0)")
                ax2.set_xlabel("y [mm]")
                ax2.grid(True, linestyle=":", linewidth=0.5)
                ax2.tick_params(direction="in", top=True, right=True)

                plt.tight_layout(rect=[0, 0, 1, 0.95])
                plt.show()

# ---------------------------------------------------------------------------
# Power density
# ---------------------------------------------------------------------------

def plot_power_density(pwr: dict, cuts: bool = True, **kwargs) -> None:
    """
    Plot power density maps from a power dictionary.

    For each polarisation in the dictionary, this plots:
        - If cuts=True: 2D power map + horizontal (y=0) and vertical (x=0) cuts.
        - If cuts=False: Only the 2D power map.

    Args:
        pwr (dict): Dictionary returned by `write_power_density` or `read_power_density`.
        cuts (bool): Whether to include 1D cuts in the plot.
        **kwargs: Optional keyword arguments, e.g., scaling factor `k`, vmin, vmax.
    """
    k = kwargs.get('k', 1)
    vmin = kwargs.get('vmin', None)
    vmax = kwargs.get('vmax', None)

    start_plotting(k)

    x = pwr['axis']['x'] * 1e3  # mm
    y = pwr['axis']['y'] * 1e3  # mm
    X, Y = np.meshgrid(x, y)

    fctr = (x[-1] - x[0]) / (y[-1] - y[0])

    for pol in [p for p in pwr if p != 'axis']:
        data = pwr[pol]['map']
        integrated = pwr[pol]['integrated']
        peak = pwr[pol]['peak']

        fig = plt.figure(figsize=(4.2 * fctr, 4))
        fig.suptitle(f"({pol}) | power: {integrated:.3e} W | peak: {peak:.2f} W/mm²",
                     fontsize=16 * k, x=0.5)

        ax = fig.add_subplot(111)
        im = ax.pcolormesh(X, Y, data, shading='auto', cmap='plasma', vmin=vmin, vmax=vmax)
        ax.set_aspect('equal')
        ax.set_xlabel('x [mm]')
        ax.set_ylabel('y [mm]')
        ax.grid(True, linestyle=':', linewidth=0.5)
        cb = plt.colorbar(im, ax=ax, fraction=0.046, pad=0.04)#, format='%.1e')
        plt.show()

        if cuts:
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 4), sharey=True)
            ix0 = np.argmin(np.abs(x))
            iy0 = np.argmin(np.abs(y))

            ax1.plot(x, data[iy0, :], color='darkred', lw=1.5)
            ax1.set_title('Hor. cut (y=0)')
            ax1.set_xlabel('x [mm]')
            ax1.set_ylabel('power density [W/mm²]')
            ax1.grid(True, linestyle=':', linewidth=0.5)
            ax1.tick_params(direction='in', top=True, right=True)

            ax2.plot(y, data[:, ix0], color='darkred', lw=1.5)
            ax2.set_title('Ver. cut (x=0)')
            ax2.set_xlabel('y [mm]')
            ax2.grid(True, linestyle=':', linewidth=0.5)
            ax2.tick_params(direction='in', top=True, right=True)

            plt.tight_layout(rect=[0, 0, 1, 0.95])
            plt.show()

# ---------------------------------------------------------------------------
# Cross-spectral density
# ---------------------------------------------------------------------------

def plot_csd(cmdDict: dict, cuts: bool = True, **kwargs) -> None:
    """
    Plot cross-spectral density (CSD) and occupation/cumulative plots from a coherent mode decomposition dictionary.

    For each direction ('H', 'V'), this plots:
        - The 2D CSD map with optional horizontal and vertical cuts.
        - The eigenmode occupation vs mode number, with cumulative sum on a secondary y-axis.

    Args:
        cmdDict (dict): CMD dictionary as returned by write_cmd or read_cmd.
        cuts (bool, optional): Whether to include 1D cuts in the CSD plots (default: True).
        **kwargs:
            k (float, optional): Scaling factor for fonts and titles (default: 1).
            nmode_max (int, optional): Maximum mode number to plot. If None, plot all modes.
    """
    k = kwargs.get('k', 1)
    nmode_max = kwargs.get('nmodes', None)
    start_plotting(k)

    for direction, data in cmdDict['source'].items():
        axis = data['axis'] * 1e3  # convert to mm
        CSD = data['CSD']
        occupation = data['occupation']
        cumulative = data['cumulated']
        CF = data['CF']

        X, Y = np.meshgrid(axis, axis)

        fctr = (axis[-1] - axis[0]) / (axis[-1] - axis[0])

        # CSD 2D plot
        fig = plt.figure(figsize=(4.2 * fctr, 4))
        fig.suptitle(f"({direction}) | cross spectral density", fontsize=16 * k, x=0.5)
        ax = fig.add_subplot(111)
        im = ax.pcolormesh(X, Y, CSD, shading='auto', cmap='terrain')
        ax.set_aspect('equal')
        ax.set_xlabel(f'{direction.lower()}$_1$ [mm]')
        ax.set_ylabel(f'{direction.lower()}$_2$ [mm]')
        ax.grid(True, linestyle=':', linewidth=0.5)
        cb = plt.colorbar(im, ax=ax, fraction=0.046, pad=0.04, format='%.0e')
        plt.show()

        if cuts:
            ix0 = np.argmin(np.abs(axis))
            iy0 = np.argmin(np.abs(axis))
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 4), sharey=True)

            ax1.plot(axis, CSD[iy0, :], color='darkred', lw=1.5)
            ax1.set_title('Hor. cut (y=0)')
            ax1.set_xlabel(f'{direction.lower()}$_1$ [mm]')
            ax1.set_ylabel('CSD')
            ax1.grid(True, linestyle=':', linewidth=0.5)
            ax1.tick_params(direction='in', top=True, right=True)
            
            ax2.plot(axis, CSD[:, ix0], color='darkred', lw=1.5)
            ax2.set_title('Ver. cut (x=0)')
            ax2.set_xlabel(f'{direction.lower()}$_2$ [mm]')
            ax2.grid(True, linestyle=':', linewidth=0.5)
            ax2.tick_params(direction='in', top=True, right=True)

            plt.tight_layout(rect=[0, 0, 1, 0.95])
            plt.show()

        # occupation vs cumulative plot
        mode_numbers = np.arange(1, len(data['eigenvalues']) + 1)

        if nmode_max is not None:
            mode_numbers = mode_numbers[:nmode_max]
            occupation = occupation[:nmode_max]
            cumulative = cumulative[:nmode_max]

        fig, ax1 = plt.subplots(figsize=(8, 4))
        color1 = 'darkred'
        ax1.plot(mode_numbers, occupation, color=color1, marker='o', linestyle='--', 
                 markerfacecolor='white', markeredgecolor=color1, markeredgewidth=1.5, 
                 lw=1.5, label='Occupation')
        ax1.set_xlabel('mode number')
        ax1.set_ylabel('occupation', color=color1)
        ax1.set_ylim(bottom=-0.05, top=1.05)
        ax1.tick_params(axis='y', labelcolor=color1)
        ax1.grid(True, linestyle=':', linewidth=0.5)
        ax1.tick_params(direction='in', top=True, right=True)

        ax2 = ax1.twinx()
        color2 = 'steelblue'
        ax2.plot(mode_numbers, cumulative, color=color2, marker='s', linestyle='--', 
                 markerfacecolor='white', markeredgecolor=color2, markeredgewidth=1.5,
                 lw=1.5, label='Cumulative')
        ax2.set_ylabel('cumulative', color=color2)
        ax2.tick_params(axis='y', labelcolor=color2)
        ax2.tick_params(direction='in', top=True, right=True)

        fig.suptitle(f"({direction}) | modes | CF {CF*100:.1f}%", fontsize=16 * k, x=0.5)
        plt.tight_layout()
        plt.show()


# SPDX-License-Identifier: CECILL-2.1
# Copyright (c) 2025 Synchrotron SOLEIL

"""
Public API for barc4sr
"""

from __future__ import annotations

from .core.energy import energy_wavelength
from .core.electron_beam import ElectronBeam
from .core.magnetic_structure import MagneticStructure
from .core.sources import (
    SynchrotronSource,
    BendingMagnetSource,
    ArbitraryMagnetSource,
)
from .core.magnetic_fields import (
    bm_magnetic_field,
    arb_magnetic_field,
    multi_bm_magnetic_field,
    multi_arb_magnetic_field
)

from .calculations.trajectory import electron_trajectory

from .calculations.radiation import (
    wavefront,
    power_density,
)

from . import plotting

from .io.rw import (
    write_electron_trajectory,
    read_electron_trajectory,
    write_wavefront,
    read_wavefront,
    write_power_density,
    read_power_density
)

__all__ = [
    # core
    "energy_wavelength",
    "ElectronBeam",
    "MagneticStructure",
    "SynchrotronSource",
    "BendingMagnetSource",
    "ArbitraryMagnetSource",
    "bm_magnetic_field",
    "arb_magnetic_field",
    "multi_bm_magnetic_field",
    "multi_arb_magnetic_field",

    # calculations
    "electron_trajectory",
    "wavefront",
    "power_density",

    # plotting namespace
    "plotting",

    # I/O
    "write_electron_trajectory",
    "read_electron_trajectory",
    "write_wavefront",
    "read_wavefront",
    "write_power_density",
    "read_power_density",
]



# SPDX-License-Identifier: CECILL-2.1
# Copyright (c) 2025 Synchrotron SOLEIL

"""
Electron trajectory.
"""

from __future__ import annotations

import time

from barc4sr.backend.srw.interface import set_light_source
from barc4sr.io.rw import write_electron_trajectory
from barc4sr.syned.mapping import barc4sr_dictionary, syned_dictionary
from barc4sr.utils.time import print_elapsed_time


def electron_trajectory(**kwargs) -> dict:
    """
    Calculates electron trajectory using SRW.

    Optional Args (kwargs):
        file_name (str): The name of the output file.
        json_file (optional): The path to the SYNED JSON configuration file.
        light_source (optional): barc4sr.aux_utils.SynchrotronSource or inheriting class
        verbose (bool): If True, print log information
    Returns:
        Dict: A dictionary containing arrays of photon energy and flux.
    """

    t0 = time.time()
    verbose = kwargs.get('verbose', True)
    json_file = kwargs.get('json_file', None)
    light_source = kwargs.get('light_source', None)
    file_name = kwargs.get('file_name', None)

    if json_file is None and light_source is None:
        raise ValueError("Please, provide either json_file or light_source (see function docstring)")
    if json_file is not None and light_source is not None:
        raise ValueError("Please, provide either json_file or light_source - not both (see function docstring)")

    if json_file is not None:
        bl = syned_dictionary(json_file, 100, 1e-3, 1e-3, 0, 0)
    if light_source is not None:
        bl = barc4sr_dictionary(light_source, 100, 1e-3, 1e-3, 0, 0)

    calc_etraj = True

    if bl['Class'] == 'bm':
        source_type = 'bending magnet'
    elif bl['Class'] == 'u':
        source_type = 'undulator'
    elif bl['Class'] == 'w':
        source_type = 'wiggler'
    elif bl['Class'] == 'arb':
        source_type = 'arbitrary magnetic field'
        
    function_txt = f"{source_type} electron trajectory using SRW:"

    if verbose: print(f"{function_txt} please wait...")

    ebeam_initial_condition = kwargs.get('ebeam_initial_condition', 6*[0])

    eBeam, magFldCnt, eTraj = set_light_source(bl,
                                               calc_etraj,
                                               ebeam_initial_condition,
                                               verbose)

    if verbose: print('completed')
    if verbose: print_elapsed_time(t0)

    eTrajDict = write_electron_trajectory(file_name, eTraj)

    return eTrajDict


# SPDX-License-Identifier: CECILL-2.1
# Copyright (c) 2025 Synchrotron SOLEIL

"""
Unified public plotting API for barc4sr.
"""

from __future__ import annotations

from .trajectory import (
    plot_electron_trajectory,
    plot_magnetic_field,
    plot_field_and_twiss,
)

from .spatial import (
    plot_wavefront,
    plot_power_density,
    plot_csd,
)

from .spectral import (
    plot_spectrum,
    plot_multiple_spectra,
    plot_beamline_acceptance_scan,
)

__all__ = [
    # trajectory
    "plot_electron_trajectory",
    "plot_magnetic_field",
    "plot_field_and_twiss",

    # spatial
    "plot_wavefront",
    "plot_power_density",
    "plot_csd",

    # spectral
    "plot_spectrum",
    "plot_multiple_spectra",
    "plot_beamline_acceptance_scan",
]


# SPDX-License-Identifier: CECILL-2.1
# Copyright (c) 2025 Synchrotron SOLEIL

"""
Synchrotron light sources built from ElectronBeam and MagneticStructure.
"""

from __future__ import annotations

from copy import deepcopy

import numpy as np
from scipy.constants import physical_constants

from barc4sr.syned.mapping import write_syned_file

from .electron_beam import ElectronBeam
from .magnetic_structure import MagneticStructure

CHARGE = physical_constants["atomic unit of charge"][0]
LIGHT = physical_constants["speed of light in vacuum"][0]
MASS = physical_constants["electron mass"][0]
PLANCK = physical_constants["Planck constant"][0]

class SynchrotronSource(object):
    """
    Base container for a synchrotron radiation sources.

    A SynchrotronSource combines an ElectronBeam and a MagneticStructure. 
    """

    def __init__(self, electron_beam: ElectronBeam, magnetic_structure: MagneticStructure) -> None:
        """
        Initialize a synchrotron source container

        Parameters
        ----------
        - electron_beam (ElectronBeam): an isntance of ElectronBeam
        - magnetic_structure (MagneticStructure): an isntance of MagneticStructure
        """
        self.ElectronBeam = electron_beam
        self.MagneticStructure = magnetic_structure

    def __getattr__(self, name: str):
        """
        Retrieves an attribute from either the ElectronBeam or MagneticStructure instances if it exists.

        Parameters
        ----------
        - name (str): Name of the requested attribute.

        Returns
        -------
        - Value of the attribute if it exists on ElectronBeam or MagneticStructure.

        Raises
        ------
        AttributeError
            If the attribute is not found on the source, the electron
            beam, or the magnetic structure.
        """
        if name in self.__dict__:
            return self.__dict__[name]
        if hasattr(self.ElectronBeam, name):
            return getattr(self.ElectronBeam, name)
        if hasattr(self.MagneticStructure, name):
            return getattr(self.MagneticStructure, name)
        raise AttributeError(f"'SynchrotronSource' object has no attribute '{name}'")

    def write_syned_config(self, json_file: str, light_source_name: str = None) -> None:
        """
        Write a SYNED JSON configuration file.

        Parameters
        ----------
        - json_file (str): The path to the JSON file where the dictionary will be written.
        - light_source_name (str): The name of the light source.
        """
        if light_source_name is None:
            light_source_name = json_file.split("/")[-1].replace(".json", "")

        write_syned_file(json_file, light_source_name, self.ElectronBeam, self.MagneticStructure)

    def print_attributes(self) -> None:
        """
        Print attributes of the electron beam and magnetic structure.
        """
        self.ElectronBeam.print_attributes()
        self.MagneticStructure.print_attributes()

# ------------------------------------------------------------------
# Arbitrary/measured magnetic field
# ------------------------------------------------------------------

class ArbitraryMagnetSource(SynchrotronSource):
    """
    SR source based on a user-defined arbitrary magnetic field.
    """

    def __init__(self, **kwargs) -> None:
        """
        Initialize an arbitrary magnetic source.

        Parameters
        ----------
        - electron_beam (ElectronBeam)
        - magnetic_structure (MagneticStructure)
        """
        super().__init__(**kwargs)

        if getattr(self.MagneticStructure, "magnet_type", None) != "arbitrary":
            raise ValueError(
                "ArbitraryMagnetSource requires MagneticStructure with "
                "magnet_type='arbitrary'."
            )
        self._original_magnetic_field = None

    def configure(self, *, si=-1e23, sf=1e23, dS=None, reset=False, verbose=False) -> None:
        """
        Configure an arbitrary magnetic source.

        Trim and optionally recenter the arbitrary magnetic field.      

        Parameters
        ----------
        - si, sf (float): lower and upper trimming bounds [m]. 
        - dS (float): shift applied to the magnetic-field grid [m].
        - reset (bool): if True, restore the original untrimmed field before
            applying dS and trimming.
        - verbose (bool): if True, prints to the prompt

        Raises
        ------
        ValueError
            If trimming removes all samples, or if the magnetic field
            dictionary is missing mandatory keys.
        """

        self._ensure_original_field()

        if reset:
            self.reset_field()

        if dS is not None:
            self.recenter(dS=dS, verbose=False)
        self.trim(si=si, sf=sf, verbose=False)

        if verbose:
            self._verbose()

    def recenter(self, *, dS=None, verbose=False) -> None:
        """
        Recenter the arbitrary magnetic field.

        The recentering is performed on the current magnetic field stored in
        MagneticStructure.magnetic_field and does not crop any data.

        Parameters
        ----------
        - dS (float): shift applied to the magnetic-field grid [m]. If None, the midpoint
            of the current s-interval is used.
        - verbose (bool): if True, prints to the prompt
        """
        self._ensure_original_field()

        mf = self.MagneticStructure.magnetic_field
        if mf is None:
            raise ValueError(
                "magnetic_field dictionary is not set in MagneticStructure "
                "(magnet_type='arbitrary' expected)."
            )

        try:
            s = np.asarray(mf["s"], dtype=float)
        except KeyError as exc:
            raise ValueError(
                "The magnetic_field dictionary must contain key 's'."
            ) from exc

        if s.ndim != 1:
            raise ValueError("'s' must be a 1D array.")

        mid = 0.5 * (s[0] + s[-1])
        center_val = mid if dS is None else float(dS)
        delta = center_val - mid

        mf_shift = {}
        for k, v in mf.items():
            if k == "s":
                mf_shift["s"] = s + delta
            else:
                mf_shift[k] = v

        self.MagneticStructure.magnetic_field = mf_shift
        self.MagneticStructure.center = center_val

        if verbose:
            self._verbose()

    def trim(self, *, si=-1e23, sf=1e23, verbose=False) -> None:
        """
        Trim the arbitrary magnetic field.

        The trimming is performed on the current magnetic field stored in
        MagneticStructure.magnetic_field. If the field has been recentered
        (via recenter or configure), the trimming bounds si and sf are
        interpreted in that recentered coordinate system.

        Parameters
        ----------
        - si, sf (float): lower and upper trimming bounds [m] in the
            current coordinate system (usually recentered).
        - verbose (bool): if True, prints to the prompt

        Raises
        ------
        ValueError
            If trimming removes all samples, or if the magnetic field
            dictionary is missing mandatory keys.
        """
        self._ensure_original_field()

        mf = self.MagneticStructure.magnetic_field
        if mf is None:
            raise ValueError(
                "magnetic_field dictionary is not set in MagneticStructure "
                "(magnet_type='arbitrary' expected)."
            )

        try:
            s = np.asarray(mf["s"], dtype=float)
            B = np.asarray(mf["B"], dtype=float)
        except KeyError as exc:
            raise ValueError(
                "The magnetic_field dictionary must contain keys 's' and 'B'."
            ) from exc

        if s.ndim != 1:
            raise ValueError("'s' must be a 1D array.")
        if B.ndim != 2 or B.shape[1] != 3 or B.shape[0] != s.size:
            raise ValueError(
                "'B' must be a (N, 3) array and its first dimension must "
                "match len(s)."
            )

        if si > sf:
            si, sf = sf, si

        mask = (s >= si) & (s <= sf)
        if not np.any(mask):
            raise ValueError(
                f"Trimming interval [{si}, {sf}] produced an empty dataset."
            )

        s_trim = s[mask]

        mf_trim = {}
        for k, v in mf.items():
            if k == "s":
                mf_trim["s"] = s_trim
                continue
            arr = np.asarray(v)
            if arr.shape[:1] == (s.size,):
                mf_trim[k] = arr[mask, ...]
            else:
                mf_trim[k] = v

        self.MagneticStructure.magnetic_field = mf_trim
        self.MagneticStructure.center = (mf_trim["s"][-1]+mf_trim["s"][0])/2

        if verbose:
            self._verbose()

    def reset_field(self) -> None:
        """
        Reset the magnetic field to its original untrimmed state.
        """
        if self._original_magnetic_field is None:
            raise RuntimeError(
                "reset_field() called before configure(). "
                "There is no stored original magnetic field to restore."
            )

        self.MagneticStructure.magnetic_field = deepcopy(self._original_magnetic_field)
        self.MagneticStructure.center = 0.0

    def _ensure_original_field(self) -> None:
        """
        Internal helper to cache the original magnetic field.

        Makes a deepcopy of MagneticStructure.magnetic_field on first use.
        """
        if self._original_magnetic_field is None:
            if self.MagneticStructure.magnetic_field is None:
                raise ValueError(
                    "magnetic_field dictionary is not set in MagneticStructure "
                    "(magnet_type='arbitrary' expected)."
                )
            self._original_magnetic_field = deepcopy(self.MagneticStructure.magnetic_field)

    def _verbose(self) -> None:
        """
        Internal helper print info on current magnetic field.
        """    
        s_out = self.MagneticStructure.magnetic_field["s"]
        print('\n>>>>>>>>>>> User-defined arbitrary magnetic field <<<<<<<<<<<\n')
        print(f"\t>> Field (re)centered at s = {self.MagneticStructure.center:.6f} m")
        print(f"\t>> Span: [{s_out[0]:.6f}, {s_out[-1]:.6f}] m")
        if s_out.size > 1:
            print(f"\t>> Step size: {s_out[1] - s_out[0]:.6g} m; N = {s_out.size}")
        else:
            print("Only one sample (no trimming).")

# ------------------------------------------------------------------
# Bending magnet
# ------------------------------------------------------------------

class BendingMagnetSource(SynchrotronSource):
    """
    SR bending magnet source
    """

    def __init__(self, **kwargs) -> None:
        """
        Initialize a bending magnet source.

        Parameters
        ----------
        - electron_beam (ElectronBeam)
        - magnetic_structure (MagneticStructure)
        """
        super().__init__(**kwargs)

        if getattr(self.MagneticStructure, "magnet_type", None) != "bending_magnet":
            raise ValueError(
                "BendingMagnetSource requires MagneticStructure with "
                "magnet_type='bending_magnet'."
            )

    def configure(self,
        *,
        center: float = None,
        extraction_angle: float = None,
        verbose: bool = False,
    ) -> None:
        """
        Configure bending magnet source.

        From B and the electron beam energy, the radius and critical energy can be computed
        on the fly and optionally printed.

        The extraction geometry is defined by either the center position [m] of the magnetic
        field or the extraction angle [rad] taken from the preceeding straight section. 

        Parameters
        ----------
        - center (float): center of the magnetic field [m]. The emission direction (optical
            axis) is taken from the tangent to the trajectory at this position. 
        - extraction_angle (float): extraction angle [rad] taken from the previous straight
            section. 0 [rad] is the magnet entrance and L/R [rad] is the magnet exit. If 
            provided, the center is derived from it and stored in MagneticStructure.center.
        - verbose (bool): if True, prints to the prompt

        Raises
        ------
        ValueError
            If B is not set, or if both center and extraction_angle are
            provided at the same time.
        """
        _ = self.B

        if verbose:
            print('\n>>>>>>>>>>> bending magnet <<<<<<<<<<<\n')
            print("> Properties:")
            print(f"\t>> B = {self.B:.6f} T")
            print(f"\t>> R = {self.radius:.6f} m")
            print(f"\t>> critical energy = {self.critical_energy:.3f} eV")

        if center is not None and extraction_angle is not None:
            raise ValueError("Provide only one of 'center' or 'extraction_angle', not both.")

        if center is None and extraction_angle is None:
            if self.MagneticStructure.center is not None:
                center = self.MagneticStructure.center
            elif self.MagneticStructure.extraction_angle is not None:
                extraction_angle = self.MagneticStructure.extraction_angle
            else:
                center = 0.0

        self._set_geometry(center=center, extraction_angle=extraction_angle, verbose=verbose)

    def _set_geometry(self, *,
        center: float, extraction_angle: float,
        verbose: bool = False,
    ) -> None:
        """
        Internal helper to set extraction geometry.

        Parameters
        ----------
        - center (float): center of the magnetic field [m]. The emission direction (optical
            axis) is taken from the tangent to the trajectory at this position. 
        - extraction_angle (float): extraction angle [rad] taken from the previous straight
            section. 0 [rad] is the magnet entrance and L/R [rad] is the magnet exit. If 
            provided, the center is derived from it and stored in MagneticStructure.center.
        - verbose (bool): if True, prints to the prompt

        Raises
        ------
        ValueError
            If both center and extraction_angle are None, or if the
            resulting angle is out of range.
        """
        if center is None and extraction_angle is None:
            raise ValueError("Provide 'center' (m) or 'extraction_angle' (rad).")

        L = self.MagneticStructure.field_length
        if L is None:
            raise ValueError("MagneticStructure.field_length must be defined.")

        R = self.radius

        half_arc = 0.5 * L / R
        full_arc = L / R

        if extraction_angle is None:
            center_val = float(center)
            extraction_angle = half_arc - center_val / R
        else:
            extraction_angle = float(extraction_angle)
            center_val = (half_arc - extraction_angle) * R

        if not (0.0 <= extraction_angle <= full_arc):
            raise ValueError(
                f"extraction_angle={extraction_angle:.6g} rad out of range [0, {full_arc:.6g}]"
            )

        magnetic_field_center = 0.5 * L - center_val

        self.MagneticStructure.center = center_val
        self.MagneticStructure.extraction_angle = extraction_angle

        if verbose:
            print("> Extraction geometry:")
            print(f"\t>> dist. from BM center    : {center_val:.6f} m")
            print(f"\t>> dist. from BM entrance  : {magnetic_field_center:.6f} m")
            print(f"\t>> extraction angle        : {extraction_angle*1e3:.3f} mrad")
            print(f"\t>> BM arc                  : {2.0*half_arc*1e3:.3f} mrad "
                  f"(L={L:.3f} m, R={R:.3f} m)")

    @property
    def B(self) -> float:
        """
        Magnetic field amplitude [T].
        """
        B = self.MagneticStructure.B
        if B is None:
            raise ValueError("Magnetic field B is not defined in MagneticStructure.")
        return B

    @B.setter
    def B(self, value: float) -> None:
        """
        Set magnetic field amplitude [T].
        """
        self.MagneticStructure.B = value

    @property
    def center(self) -> float:
        """
        Center of the magnetic field [m].
        """
        return self.MagneticStructure.center

    @center.setter
    def center(self, value: float) -> None:
        """
        Set center of the magnetic field [m].
        """
        self.MagneticStructure.center = value

    @property
    def extraction_angle(self) -> float:
        """
        Extraction angle [rad] taken from the previous straight section.
        """
        return self.MagneticStructure.extraction_angle

    @extraction_angle.setter
    def extraction_angle(self, value: float) -> None:
        """
        Set extraction angle [rad] taken from the previous straight section.
        """
        self.MagneticStructure.extraction_angle = value

    @property
    def radius(self) -> float:
        """
        Bending radius [m], computed from B and electron beam energy.
        """
        B = self.B
        gamma = self.gamma()
        e_speed = LIGHT * np.sqrt(1.0 - 1.0 / gamma**2)
        return gamma * MASS * e_speed / (CHARGE * abs(B))

    @property
    def critical_energy(self) -> float:
        """
        Critical photon energy [eV], computed from B and electron beam energy.
        """
        B = self.B
        gamma = self.gamma()
        return (3.0 * PLANCK * B * gamma**2) / (4.0 * np.pi * MASS)
    
# ------------------------------------------------------------------
# Undulator
# ------------------------------------------------------------------

class UndulatorSource(SynchrotronSource):
    """
    Placeholder: SR Undulator source.
    """

    def __init__(self, **kwargs) -> None:
        """
        Initialize a undulator source.

        Parameters
        ----------
        - electron_beam (ElectronBeam)
        - magnetic_structure (MagneticStructure)
        """
        super().__init__(**kwargs)

        if getattr(self.MagneticStructure, "magnet_type", None) != "undulator":
            raise ValueError(
                "WigglerSource requires a MagneticStructure with "
                "magnet_type='undulator'."
            )

    def initialize(self, **kwargs) -> None:
        """
        Configure the undulator source.

        Parameters
        ----------
        **kwargs
            Reserved for future configuration options.

        Raises
        ------
        NotImplementedError
            Always raised until the undulator model is implemented.
        """
        raise NotImplementedError(
            "WigglerSource.initialize() is not implemented yet. "
            "Please implement the undulator radiation model."
        )

# ------------------------------------------------------------------
# Wiggler
# ------------------------------------------------------------------

class WigglerSource(SynchrotronSource):
    """
    Placeholder: SR Wiggler source.
    """

    def __init__(self, **kwargs) -> None:
        """
        Initialize a wiggler source.

        Parameters
        ----------
        - electron_beam (ElectronBeam)
        - magnetic_structure (MagneticStructure)
        """
        super().__init__(**kwargs)

        if getattr(self.MagneticStructure, "magnet_type", None) != "wiggler":
            raise ValueError(
                "WigglerSource requires a MagneticStructure with "
                "magnet_type='wiggler'."
            )

    def initialize(self, **kwargs) -> None:
        """
        Configure the wiggler source.

        Parameters
        ----------
        **kwargs
            Reserved for future configuration options.

        Raises
        ------
        NotImplementedError
            Always raised until the wiggler model is implemented.
        """
        raise NotImplementedError(
            "WigglerSource.initialize() is not implemented yet. "
            "Please implement the wiggler radiation model."
        )


# SPDX-License-Identifier: CECILL-2.1
# Copyright (c) 2025 Synchrotron SOLEIL

"""
_version.py - library version
"""

from importlib.metadata import PackageNotFoundError, version

__all__ = ["__version__"]

try:
    __version__ = version("barc4sr")
except PackageNotFoundError:

    try:
        import tomllib
        from pathlib import Path

        root = Path(__file__).resolve().parents[2]
        data = tomllib.loads((root / "pyproject.toml").read_text(encoding="utf-8"))
        __version__ = data.get("project", {}).get("version", "0+unknown")
    except Exception:
        __version__ = "0+unknown"